package demoplugin.cbp;

import org.zeroturnaround.bundled.javassist.ClassPool;
import org.zeroturnaround.bundled.javassist.CtClass;
import org.zeroturnaround.bundled.javassist.CtMethod;
import org.zeroturnaround.javarebel.integration.support.JavassistClassBytecodeProcessor;

import demoplugin.ReloadableServlet;

/**
 * MemoryUsageServletCBP a.k.a. MemoryUsageServlet Class Bytecode Processor.
 * <p>
 *   This is used to patch the MemoryUsageServlet at load time.
 *   The patch does the following:
 *   <ul>
 *     <li>adds a method in MemoryUsageServlet for reloading the properties</li>
 *     <li>adds a method in MemoryUsageServlet for reinitializing the memoryUsageMonitor</li>
 *     <li>makes MemoryUsageServlet implement the ReloadableServlet interface</li>
 *     <li>registers JRebel event listeners at the end of the init method</li>
 *   </ul>
 * </p>
 * <p>
 *   The patch is written using a bundled version of the Javassist bytecode manipulation library.
 *   Please read the official documentation at https://github.com/jboss-javassist/javassist/wiki for details.
 * </p>
 */
public class MemoryUsageServletCBP extends JavassistClassBytecodeProcessor {

  @Override
  public void process(ClassPool cp, ClassLoader cl, CtClass ctClass) throws Exception {
    // ctClass - javassist representation of the MemoryUsageServlet class
    // we can change the MemoryUsageServlet class here (before it's loaded by the classloader)

    // importPackage works like the import statements at the beginning of a regular .java file.
    cp.importPackage("demoapp");
    cp.importPackage("demoplugin");
    cp.importPackage("java.io");

    // add a new method in the MemoryUsageServlet class for reloading the properties
    ctClass.addMethod(CtMethod.make("" +
        "public void reloadProperties() throws IOException {" +
        "  properties = loadProperties();" +
        "}", ctClass));

    // add a new method in the MemoryUsageServlet class for reinitializing the memoryUsageMonitor
    ctClass.addMethod(CtMethod.make("" +
        "public void reloadMonitor() {" +
        "  if (memoryUsageMonitor != null) {" +
        "    memoryUsageMonitor.close();" +
        "  }" +
        "  memoryUsageMonitor = new MemoryUsageMonitor(" +
        "      ApplicationConfiguration.getMaxSamples()," +
        "      ApplicationConfiguration.getSampleDelayMillis());" +
        "}", ctClass));

    // instead you could add the reload* methods directly into the MemoryUsageServlet class;
    // it's more intrusive, but this way you won't have to patch them in

    // we want to call the added reload* methods from our JRebel event listeners. since the
    // methods are only added at load time, we cannot use them in our code directly (compiler
    // can't find the methods). as a workaround, we'll create the ReloadableServlet interface.
    // we declare ReloadableServlet in our JRebel plugin and add the new method signatures in it.
    // then we patch the MemoryUsageServlet class to implement the interface. we can now use the
    // interface to call the methods from MemoryUsageServlet that are added at load time.

    // class MemoryUsageServlet implements ReloadableServlet
    ctClass.addInterface(cp.get(ReloadableServlet.class.getName()));

    // insert the code snippet at the end of the MemoryUsageServlet#init method.
    // note that ReloadHelper methods use ReloadableServlet as the first argument.
    ctClass.getDeclaredMethod("init").insertAfter("" +
        "ReloadHelper.addRequestListener(this);" +
        "ReloadHelper.addClassChangeListener(this, ApplicationConfiguration.class);");
  }
}
